from __future__ import annotations

import altair as alt
import pandas as pd
from pytest import fixture, raises

import larch as lx


@fixture(scope="module")
def model():
    m = lx.example(1)
    m.loglike(
        {
            "ASC_BIKE": -2.3776321824746316,
            "ASC_SR2": -2.1764779675270782,
            "ASC_SR3P": -3.723708490793479,
            "ASC_TRAN": -0.6671048298163328,
            "ASC_WALK": -0.2012241299742951,
            "hhinc#2": -0.0021788304992503785,
            "hhinc#3": 0.00035610619543557237,
            "hhinc#4": -0.005300245144846563,
            "hhinc#5": -0.012728753373485677,
            "hhinc#6": -0.009701442798519953,
            "totcost": -0.004917728615263069,
            "tottime": -0.051427250886585466,
        }
    )
    return m


def test_analyze_age(model, dataframe_regression):
    df = model.analyze_predictions_co("age")
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


def test_analyze_age_fig(model):
    fig = model.analyze_predictions_co_figure("age")
    assert isinstance(fig, alt.TopLevelMixin)


def test_analyze_age_binned(model, dataframe_regression):
    df = model.analyze_predictions_co("age", bins=[0, 25, 45, 65, 99])
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


def test_analyze_age_bool(model, dataframe_regression):
    df = model.analyze_predictions_co("age > 35")
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


def test_analyze_age_bool_fig(model):
    fig = model.analyze_predictions_co_figure("age > 35")
    assert isinstance(fig, alt.TopLevelMixin)


def test_analyze_few_values(model, dataframe_regression):
    df = model.analyze_predictions_co("age // 70")
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


def test_analyze_few_values_fig(model):
    fig = model.analyze_predictions_co_figure("age // 70")
    assert isinstance(fig, alt.TopLevelMixin)


def test_analyze_age_weighted(model, dataframe_regression):
    df = model.analyze_predictions_co("age", wgt="1.5 if wkccbd else 1.0")
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


def test_analyze_elasticity_hhinc(model, dataframe_regression):
    df = model.analyze_elasticity("hhinc")
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


def test_analyze_elasticity_totcost_1(model, dataframe_regression):
    df = model.analyze_elasticity("totcost", altid=1)
    assert isinstance(df, pd.io.formats.style.Styler)
    df = df.data
    dataframe_regression.check(df)


# variants that should fail
def test_analyze_elasticity_idco_with_altid(model):
    with raises(KeyError):
        # giving an altid when the variable is idco
        model.analyze_elasticity("hhinc", altid=1)


def test_analyze_elasticity_missing_name(model):
    with raises(KeyError):
        # giving a variable that is not in the model
        model.analyze_elasticity("missing_name")


def test_analyze_elasticity_expression(model):
    with raises(KeyError):
        # giving a variable is not a simple name
        model.analyze_elasticity("hhinc ** 2")


def test_analyze_elasticity_altid_as_string(model):
    with raises(KeyError):
        # giving a variable is not a simple name
        model.analyze_elasticity("totcost", altid="DA")



import logging

import numpy as np
import pandas as pd
import xarray as xr

from sharrow.dataset import Dataset

from .dataset import one_based, zero_based

logger = logging.getLogger("sharrow.translate")


def omx_to_zarr(
    omx_filenames,
    zarr_directory,
    index_names=("otaz", "dtaz", "time_period"),
    indexes=None,
    *,
    time_periods=None,
    time_period_sep="__",
):
    try:
        from larch import OMX
    except ImportError:
        raise ImportError("larch is required to read OMX files") from None

    bucket = {}

    r1 = r2 = None

    for omx_filename in omx_filenames:
        logger.info(f"reading metadata from {omx_filename}")

        omx = OMX(omx_filename)
        omx_data = omx.data
        omx_shape = omx.shape
        omx_lookup = omx.lookup

        data_names = list(omx_data._v_children.keys())
        n1, n2 = omx_shape
        if indexes is None:
            # default reads mapping if only one lookup is included, otherwise one-based
            if len(omx_lookup._v_children) == 1:
                ranger = None
            else:
                ranger = one_based
        elif indexes == "one-based":
            ranger = one_based
        elif indexes == "zero-based":
            ranger = zero_based
        elif indexes in set(omx_lookup._v_children):
            ranger = None
        else:
            raise NotImplementedError(
                "only one-based, zero-based, and named indexes are implemented"
            )
        if ranger is not None:
            r1 = ranger(n1)
            r2 = ranger(n2)
        else:
            r1 = r2 = pd.Index(omx_lookup[indexes])

        if time_periods is None:
            raise ValueError("must give time periods explicitly")

        bucket.update({i: omx.data[i] for i in omx.data._v_children})

    data_names = list(bucket.keys())

    logger.info(f"writing to {zarr_directory}")
    for k in data_names:
        logger.info(f" - {k}")
        ds = Dataset()
        if time_period_sep in k:
            base_k, time_k = k.split(time_period_sep, 1)
            ds[base_k] = xr.DataArray(
                np.float32(0),
                dims=index_names,
                coords={
                    index_names[0]: r1,
                    index_names[1]: r2,
                    index_names[2]: time_periods,
                },
            )
            ds[base_k].loc[:, :, time_k] = bucket[k][:]
        else:
            ds[k] = xr.DataArray(
                bucket[k][:],
                dims=index_names[:2],
                coords={
                    index_names[0]: r1,
                    index_names[1]: r2,
                },
            )
        ds.to_zarr(zarr_directory, mode="a")



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import MutableSequence

from .._optional import jax, jnp, js
from .param_core import ParameterBucket


class Mixture(ABC):
    """
    A mixing distribution for a model parameter.

    This is an abstract base class. Subclasses must implement the `param_names`,
    `prep`, `roll`, and `to_dict` methods.
    """

    def __init__(self):
        self._parent = None

    @abstractmethod
    def param_names(self):
        """
        Named parameters referenced by this mixture, and their default values.

        Returns
        -------
        dict
        """
        raise NotImplementedError()

    @abstractmethod
    def prep(self, bucket: ParameterBucket):
        raise NotImplementedError()

    @abstractmethod
    def roll(self, draws: jax.Array, parameters: jax.Array) -> jax.Array:
        """
        Apply this mixing distribution to some random draws.

        Parameters
        ----------
        draws : jax.Array, shape [...]
            A set of pseudo-random draws, nominally uniformly distributed
            in the range 0 to 1.
        parameters : jax.Array, shape [..., n_params]
            An array of parameters, previously broadcasted to the same shape
            as the draws, plus the parameter dimension itself.

        Returns
        -------
        parameters : jax.Array, shape [..., n_params]
            The computed distribution of the target parameter has been overlaid.
        """
        raise NotImplementedError()

    @abstractmethod
    def to_dict(self):
        raise NotImplementedError()


class MixtureList(MutableSequence):
    def __init__(self, init=None):
        self._parent = None
        self._mixtures = list()
        if init is not None:
            for i in init:
                if isinstance(i, Mixture):
                    self._mixtures.append(i)
                else:
                    raise TypeError(
                        f"members of {self.__class__.__name__} must be Mixture"
                    )

    def set_parent(self, instance):
        self._parent = instance
        for i in self._mixtures:
            i._parent = instance

    def __fresh(self, instance):
        newself = MixtureList()
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __mangle(self):
        try:
            self._parent.mangle()
        except AttributeError:
            pass

    def __get__(self, instance, owner):
        if instance is None:
            return self
        newself = getattr(instance, self.private_name, None)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, values):
        newself = getattr(instance, self.private_name, None)
        if newself is None:
            newself = self.__fresh(instance)
        else:
            newself._mixtures.clear()
        newself.__init__(values)
        newself.set_parent(instance)
        newself.__mangle()

    def __delete__(self, instance):
        newself = getattr(instance, self.private_name, None)
        if newself is not None and len(newself):
            newself.__mangle()
        if newself is None:
            newself = self.__fresh(instance)
        else:
            newself._mixtures.clear()
        newself.__init__()

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_private_" + name

    def __getitem__(self, item):
        return self._mixtures[item]

    def __setitem__(self, key: int, value):
        if not isinstance(value, Mixture):
            raise TypeError("items must be of type Mixture")
        self._mixtures[key] = value
        self.__mangle()
        self._mixtures[key]._parent = self._parent

    def __delitem__(self, key):
        del self._mixtures[key]
        self.__mangle()

    def __len__(self):
        return len(self._mixtures)

    def insert(self, index, value):
        if not isinstance(value, Mixture):
            raise TypeError("items must be of type Mixture")
        self._mixtures.insert(index, value)
        self.__mangle()
        self._mixtures[index]._parent = self._parent

    def __repr__(self):
        return repr(self._mixtures)

    def _is_duplicate(self, value):
        for i in self._mixtures:
            if i == value:
                return True
        return False

    def to_list(self):
        return [i.to_dict() for i in self._mixtures]

    def from_list(self, j):
        self._mixtures.clear()
        for i in j:
            kind = i.pop("type")
            if kind is None:
                raise ValueError("missing mixture type")
            cls = globals()[kind]
            self._mixtures.append(cls(**i))


class Normal(Mixture):
    """A normal distribution applied to a model parameter."""

    def __init__(self, mean: str, std: str):
        super().__init__()
        self.mean_ = mean
        self.std_ = std
        self.imean = -1
        self.istd = -1
        self.default_mean = 0.0
        self.default_std = 0.001

    def __repr__(self):
        return f"{self.__class__.__name__}({self.mean_!r}, {self.std_!r})"

    def __eq__(self, other):
        return (
            isinstance(other, Normal)
            and self.mean_ == other.mean_
            and self.std_ == other.std_
        )

    def param_names(self):
        return {
            self.mean_: self.default_mean,
            self.std_: self.default_std,
        }

    def prep(self, bucket: ParameterBucket):
        self.imean = bucket.get_param_loc(self.mean_)
        self.istd = bucket.get_param_loc(self.std_)

    def roll(self, draw_vec, parameters):
        assert self.imean >= 0
        assert self.istd >= 0
        v = js.stats.norm.ppf(
            draw_vec, parameters[..., self.imean], parameters[..., self.istd]
        )
        parameters = parameters.at[..., self.imean].set(v)
        return parameters

    def to_dict(self):
        return dict(
            type=self.__class__.__name__,
            mean=self.mean_,
            std=self.std_,
        )


class LogNormal(Normal):
    """
    A log-normal distribution applied to a model parameter.

    The "mean" and "std" parameters are the mean and standard deviation of the
    underlying normal distribution, not the mean and standard deviation of the
    log-normal distribution.
    """

    def roll(self, draw_vec, parameters):
        assert self.imean >= 0
        assert self.istd >= 0
        v = js.stats.norm.ppf(
            draw_vec, parameters[..., self.imean], parameters[..., self.istd]
        )
        parameters = parameters.at[..., self.imean].set(jnp.exp(v))
        return parameters


class NegLogNormal(Normal):
    """
    The negative of a log-normal distribution applied to a model parameter.

    This is convenient when it is desired to ensure the parameter must be
    negative, as expected for coefficients associated with travel time or travel
    costs.

    The "mean" and "std" parameters are the mean and standard deviation of the
    underlying normal distribution, not the mean and standard deviation of the
    log-normal distribution.
    """

    def roll(self, draw_vec, parameters):
        assert self.imean >= 0
        assert self.istd >= 0
        v = js.stats.norm.ppf(
            draw_vec, parameters[..., self.imean], parameters[..., self.istd]
        )
        parameters = parameters.at[..., self.imean].set(-jnp.exp(v))
        return parameters


class Uniform(Mixture):
    """A uniform distribution applied to a model parameter."""

    def __init__(self, low: str, high: str):
        super().__init__()
        self.low_ = low
        self.high_ = high
        self.ilow = -1
        self.ihigh = -1
        self.default_low = 0.0
        self.default_high = 1.0

    def __repr__(self):
        return f"{self.__class__.__name__}({self.low_!r}, {self.high_!r})"

    def __eq__(self, other):
        return (
            isinstance(other, Uniform)
            and self.low_ == other.low_
            and self.high_ == other.high_
        )

    def param_names(self):
        return {
            self.low_: self.default_low,
            self.high_: self.default_high,
        }

    def prep(self, bucket: ParameterBucket):
        self.ilow = bucket.get_param_loc(self.low_)
        self.ihigh = bucket.get_param_loc(self.high_)

    def roll(self, draw_vec, parameters):
        assert self.ilow >= 0
        assert self.ihigh >= 0
        spread = draw_vec * (parameters[..., self.ihigh] - parameters[..., self.ilow])
        parameters = parameters.at[..., self.ilow].set(
            parameters[..., self.ilow] + spread
        )
        return parameters

    def to_dict(self):
        return dict(
            type=self.__class__.__name__,
            low=self.low_,
            high=self.high_,
        )



from __future__ import annotations


class TouchNotify:
    """A mixin to execute a callback when a mutate-possible command is called."""

    def __init__(self, *args, touch_callback_name="mangle", **kwargs):
        self._touch_name = touch_callback_name
        self._touch = lambda: None
        super().__init__(*args, **kwargs)

    def __set_name__(self, owner, name):
        # self : compiledmethod
        # owner : parent class that will have `self` as a member
        # name : the name of the attribute that `self` will be
        self.public_name = name
        self.private_name = "_touchable_" + name

    def __get__(self, obj, objtype=None):
        # self : TouchNotify
        # obj : instance of parent class that has `self` as a member, or None
        # objtype : class of `obj`
        result = getattr(obj, self.private_name, None)
        if result is not None:
            result.set_touch_callback(getattr(obj, self._touch_name, lambda: None))
        return result

    def __set__(self, obj, value):
        # self : TouchNotify
        # obj : instance of parent class that has `self` as a member
        # value : the new value that is trying to be assigned
        if not (isinstance(value, type(self)) or value is None):
            value = type(self)(value)
        setattr(obj, self.private_name, value)

    def __delete__(self, obj):
        # self : TouchNotify
        # obj : instance of parent class that has `self` as a member
        setattr(obj, self.private_name, None)

    def get_touch_callback(self):
        return self._touch

    def set_touch_callback(self, callback):
        if callable(callback):
            self._touch = callback
        elif callback is None:
            self._touch = lambda: None
        else:
            raise TypeError("callback must be callable")

    def del_touch_callback(self):
        self._touch = lambda: None

    def touch(self):
        self._touch()

    def __setitem__(self, key, value):
        ret = super().__setitem__(key, value)
        try:
            self.__touch()
        except AttributeError:
            pass  # on init, may not yet have _touch
        return ret

    def __setattr__(self, key, value):
        ret = super().__setattr__(key, value)
        try:
            self.__touch()
        except AttributeError:
            pass  # on init, may not yet have _touch
        return ret

    def __delitem__(self, key):
        ret = super().__delitem__(key)
        self._touch()
        return ret

    def __delattr__(self, key):
        ret = super().__delattr__(key)
        self._touch()
        return ret

    def setdefault(self, *args, **kwargs):
        ret = super().setdefault(*args, **kwargs)
        self._touch()
        return ret

    def append(self, *args, **kwargs):
        ret = super().append(*args, **kwargs)
        self._touch()
        return ret

    def extend(self, *args, **kwargs):
        ret = super().extend(*args, **kwargs)
        self._touch()
        return ret

    def update(self, *args, **kwargs):
        ret = super().update(*args, **kwargs)
        self._touch()
        return ret

    def insert(self, *args, **kwargs):
        ret = super().insert(*args, **kwargs)
        self._touch()
        return ret

    def remove(self, *args, **kwargs):
        ret = super().remove(*args, **kwargs)
        self._touch()
        return ret

    def pop(self, *args, **kwargs):
        ret = super().pop(*args, **kwargs)
        self._touch()
        return ret

    def popitem(self, *args, **kwargs):
        ret = super().popitem(*args, **kwargs)
        self._touch()
        return ret

    def clear(self, *args, **kwargs):
        ret = super().clear(*args, **kwargs)
        self._touch()
        return ret

    def __call__(self, *args, **kwargs):
        ret = super().__call__(*args, **kwargs)
        self._touch()
        return ret

    def __iadd__(self, other):
        ret = super().__iadd__(other)
        self._touch()
        return ret

    def __isub__(self, other):
        ret = super().__isub__(other)
        self._touch()
        return ret

    def __imul__(self, other):
        ret = super().__imul__(other)
        self._touch()
        return ret

    def __idiv__(self, other):
        ret = super().__idiv__(other)
        self._touch()
        return ret

    def __iand__(self, other):
        ret = super().__iand__(other)
        self._touch()
        return ret

    def __ior__(self, other):
        ret = super().__ior__(other)
        self._touch()
        return ret

    def __ixor__(self, other):
        ret = super().__ixor__(other)
        self._touch()
        return ret

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        self._touch = lambda: None



from __future__ import annotations

import numpy as np
from scipy.stats import gaussian_kde
from sklearn.neighbors import KernelDensity

# partly based on ReflectionBoundedKDE from
# https://git.ligo.org/lscsoft/pesummary/-/blob/master/pesummary/core/plots/bounded_1d_kde.py
#
# MIT License
# Copyright (c) 2018-2021 Charlie Hoy charlie.hoy@ligo.org
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def weighted_sample_std(values, weights, ddof=1.0):
    """Weighted sample standard deviation."""
    average = np.average(values, weights=weights)
    variance = np.average((values - average) ** 2, weights=weights)
    variance = variance * sum(weights) / (sum(weights) - ddof)
    return np.sqrt(variance)


class BoundedKDE:
    def __init__(
        self, x, weights=None, bw_method="scott", lb="min", ub="max", kernel="gaussian"
    ):
        if lb is None:
            self.lower_bound = None
        else:
            self.lower_bound = lb if lb != "min" else np.nanmin(x)
        if ub is None:
            self.upper_bound = None
        else:
            self.upper_bound = ub if ub != "max" else np.nanmax(x)
        if x.ndim > 1:
            x = x.ravel()
            if weights is not None:
                weights = weights.ravel()
        if weights is not None:
            x = x[weights > 0]
            weights = weights[weights > 0]
        if kernel is None:
            if x.size > 1000:
                kernel = "tophat"
            else:
                kernel = "gaussian"
        if bw_method == "scott":
            if weights is None:
                bw = x.std(ddof=1) * np.power(x.size, -1.0 / 5)
            else:
                bw = weighted_sample_std(x, weights, ddof=1.0) * np.power(
                    weights.sum(), -1.0 / 5
                )
        elif getattr(bw_method, "bandwidth", None) is not None:
            bw = bw_method.bandwidth
        elif np.isscalar(bw_method) and not isinstance(bw_method, str):
            bw = bw_method
        else:
            raise NotImplementedError(f"bw_method={bw_method}")
        self.bandwidth = bw
        self.kernel = kernel
        self.kde = KernelDensity(bandwidth=bw, kernel=self.kernel)
        self.kde.fit(x[:, np.newaxis], sample_weight=weights)

    def evaluate(self, x):
        if x.ndim > 1:
            x = x.ravel()
        pdf = np.exp(self.kde.score_samples(x[:, np.newaxis]))
        if self.lower_bound is not None:
            pdf += np.exp(
                self.kde.score_samples((2 * self.lower_bound - x)[:, np.newaxis])
            )
        if self.upper_bound is not None:
            pdf += np.exp(
                self.kde.score_samples((2 * self.upper_bound - x)[:, np.newaxis])
            )
        return pdf

    def __call__(self, x):
        if x.ndim > 1:
            x = x.ravel()
        results = self.evaluate(x)
        if self.lower_bound is not None:
            results[x < self.lower_bound] = 0.0
        if self.upper_bound is not None:
            results[x > self.upper_bound] = 0.0
        return results


class bounded_gaussian_kde(gaussian_kde):
    def __init__(self, dataset, bw_method=None, weights=None, lb="min", ub="max"):
        super().__init__(
            dataset,
            bw_method=bw_method,
            weights=weights,
        )
        self.lower_bound = lb if lb != "min" else np.nanmin(dataset)
        self.upper_bound = ub if ub != "max" else np.nanmax(dataset)

    def evaluate(self, pts):
        """Return an estimate of the density evaluated at the given points."""
        x = pts.T
        pdf = super().evaluate(pts.T)
        if self.lower_bound is not None:
            pdf += super().evaluate(2 * self.lower_bound - x)
        if self.upper_bound is not None:
            pdf += super().evaluate(2 * self.upper_bound - x)
        return pdf

    def __call__(self, pts):
        pts = np.atleast_1d(pts)
        results = self.evaluate(pts)
        if self.lower_bound is not None:
            results[pts < self.lower_bound] = 0.0
        if self.upper_bound is not None:
            results[pts > self.upper_bound] = 0.0
        return results



from __future__ import annotations

import shutil
from pathlib import Path

import pytest
import xarray as xr

from sharrow.datastore import DataStore, ReadOnlyError


def test_datasstore_checkpointing(tmp_path: Path, person_dataset):
    tm = DataStore(directory=tmp_path)
    tm["persons"] = person_dataset
    tm.make_checkpoint("init_persons")

    person_dataset["DoubleAge"] = person_dataset["Age"] * 2
    tm.update("persons", person_dataset["DoubleAge"])
    tm.make_checkpoint("annot_persons")

    tm2 = DataStore(directory=tmp_path)
    tm2.restore_checkpoint("annot_persons")
    xr.testing.assert_equal(tm2.get_dataset("persons"), person_dataset)

    tm2.restore_checkpoint("init_persons")
    assert "DoubleAge" not in tm2.get_dataset("persons")

    tm_ro = DataStore(directory=tmp_path, mode="r")
    with pytest.raises(ReadOnlyError):
        tm_ro.make_checkpoint("will-fail")


def test_datasstore_checkpointing_parquet(tmp_path: Path, person_dataset):
    tm = DataStore(directory=tmp_path, storage_format="parquet")
    tm["persons"] = person_dataset
    tm.make_checkpoint("init_persons")

    person_dataset["DoubleAge"] = person_dataset["Age"] * 2
    tm.update("persons", person_dataset["DoubleAge"])
    tm.make_checkpoint("annot_persons")

    tm2 = DataStore(directory=tmp_path)
    tm2.restore_checkpoint("annot_persons")
    xr.testing.assert_equal(tm2.get_dataset("persons"), person_dataset)

    tm2.restore_checkpoint("init_persons")
    assert "DoubleAge" not in tm2.get_dataset("persons")

    tm_ro = DataStore(directory=tmp_path, mode="r")
    with pytest.raises(ReadOnlyError):
        tm_ro.make_checkpoint("will-fail")


def test_datasstore_relationships(
    tmp_path: Path, person_dataset, household_dataset, tours_dataset
):
    pth = tmp_path.joinpath("relations")

    if pth.exists():
        shutil.rmtree(pth)

    pth.mkdir(parents=True, exist_ok=True)
    tm = DataStore(directory=pth)

    tm["persons"] = person_dataset
    tm.make_checkpoint("init_persons")

    tm["households"] = household_dataset
    tm.add_relationship("persons.household_id @ households.household_id")
    tm.make_checkpoint("init_households")

    tm["tours"] = tours_dataset
    tm.add_relationship("tours.person_id @ persons.person_id")
    tm.make_checkpoint("init_tours")

    tm.digitize_relationships()
    assert tm.relationships_are_digitized

    tm.make_checkpoint("digitized")

    tm2 = DataStore(directory=pth, mode="r")
    tm2.read_metadata("*")
    tm2.restore_checkpoint("init_households")

    assert sorted(tm2.get_dataset("persons")) == [
        "Age",
        "Income",
        "Name",
        "WorkMode",
        "household_id",
    ]

    assert sorted(tm2.get_dataset("households")) == [
        "n_cars",
    ]

    tm2.restore_checkpoint("digitized")
    assert sorted(tm2.get_dataset("persons")) == [
        "Age",
        "Income",
        "Name",
        "WorkMode",
        "digitizedOffsethousehold_id_households_household_id",
        "household_id",
    ]

    double_age = tm2.get_dataset("persons")["Age"] * 2
    with pytest.raises(ReadOnlyError):
        tm2.update("persons", double_age.rename("doubleAge"))

    with pytest.raises(ReadOnlyError):
        tm2.make_checkpoint("age-x2")

    tm.update("persons", double_age.rename("doubleAge"))
    assert sorted(tm.get_dataset("persons")) == [
        "Age",
        "Income",
        "Name",
        "WorkMode",
        "digitizedOffsethousehold_id_households_household_id",
        "doubleAge",
        "household_id",
    ]

    tm.make_checkpoint("age-x2")
    tm2.read_metadata()
    tm2.restore_checkpoint("age-x2")

    person_restored = tm2.get_dataframe("persons")
    print(person_restored.WorkMode.dtype)
    assert person_restored.WorkMode.dtype == "category"



import numpy as np
import pandas as pd
import xarray as xr

from .accessors import register_dataset_method


class _Df_Accessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __call__(self, *args, **kwargs) -> xr.Dataset:
        raise NotImplementedError()

    def df(self, df, *, append=False, mapping=None):
        """
        Extract values based on the coordinates indicated by columns of a DataFrame.

        Parameters
        ----------
        df : pandas.DataFrame or Mapping[str, array-like]
            The columns (or keys) of `df` should match the named dimensions of
            this Dataset.  The resulting extracted DataFrame will have one row
            per row of `df`, columns matching the data variables in this dataset,
            and each value is looked from the source Dataset.
        append : bool or str, default False
            Assign the results of this extraction to variables in a copy of the
            dataframe `df`.  Set to a string to make that a prefix for the variable
            names.
        mapping : dict, optional
            Apply this rename mapping to the column names before extracting data.

        Returns
        -------
        pandas.DataFrame
        """
        if mapping is not None:
            df_ = df.rename(columns=mapping)
        else:
            df_ = df
        keys = {i: df_[i] for i in self._obj.dims}
        result = self(**keys).reset_coords(drop=True).to_dataframe()
        if isinstance(df, pd.DataFrame):
            result.index = df.index
        if append:
            if isinstance(append, str):
                result = result.add_prefix(append)
            result = df.assign(**result)
        return result

    def ds(self, ds, *, append=False, mapping=None, compute=False):
        """
        Extract values based on the coordinates indicated by variables of a Dataset.

        Parameters
        ----------
        ds : xr.Dataset or Mapping[str, array-like]
            The variables (or keys) of `ds` should match the named dimensions of
            this Dataset, and should all be the same shape.  The resulting
            extracted Dataset will have that common shape, variables matching the
            data variables in this dataset, and each value is looked from the
            source Dataset.
        append : bool or str, default False
            Assign the results of this extraction to variables in a copy of the
            dataset `ds`.  Set to a string to make that a prefix for the variable
            names.
        mapping : dict, optional
            Apply this rename mapping to the variable names before extracting data.
        compute : bool, default False
            Trigger a `compute` on dask arrays in the result.

        Returns
        -------
        xarray.Dataset
        """
        if mapping is not None:
            ds_ = ds.rename(mapping)
        else:
            ds_ = ds
        keys = {i: ds_[i].data for i in self._obj.dims}
        key_dims = None
        for i in self._obj.dims:
            if key_dims is None:
                key_dims = ds_[i].dims
            else:
                assert key_dims == ds_[i].dims
        if len(key_dims) > 1:
            raise NotImplementedError(
                "dataset with more than one dimension not implemented"
            )
        else:
            result = self(**keys, _index_name=key_dims[0]).reset_coords(drop=True)
        if compute:
            result = result.compute()
        if append:
            if isinstance(append, str):
                renames = {k: f"{append}{k}" for k in result.variables}
                result = result.rename(renames)
            result = ds.assign(result)
        return result

    def _filter(
        self,
        *,
        _name=None,
        _names=None,
        _load=False,
        _index_name=None,
        _func=None,
        _raw_idxs=None,
        **idxs,
    ):
        loaders = {}
        if _index_name is None:
            _index_name = "index"
        for k, v in idxs.items():
            loaders[k] = xr.DataArray(v, dims=[_index_name])
        if _name is not None:
            assert isinstance(_name, str)
            _names = [_name]
        if _names:
            _baggage = list(self._obj.digital_encoding.baggage(_names))
            _all_names = list(_names) + _baggage
            ds = self._obj[_all_names]
        else:
            _baggage = []
            ds = self._obj
        if _load:
            ds = ds.load()
        if _func == "isel":
            # remove coordinates, we don't need them for isel
            ds_ = ds.drop_vars(ds.coords)
        else:
            ds_ = ds

        if _names:
            result = (
                getattr(ds_, _func)(**loaders)
                .digital_encoding.strip(_names)
                .drop_vars(_baggage)
            )
            for n in _names:
                if self._obj.redirection.is_blended(n):
                    dims = self._obj[n].dims
                    result[n] = xr.DataArray(
                        self._obj.redirection.get_blended(
                            n,
                            result[n].to_numpy(),
                            _raw_idxs[dims[0]],
                            _raw_idxs[dims[1]],
                        ),
                        dims=result[n].dims,
                    )
            if _name is not None:
                result = result[_name]
            return result
        else:
            result = getattr(ds_, _func)(**loaders)
            names = list(result.keys())
            for n in names:
                if self._obj.redirection.is_blended(n):
                    dims = self._obj[n].dims
                    result[n] = xr.DataArray(
                        self._obj.redirection.get_blended(
                            n,
                            result[n].to_numpy(),
                            _raw_idxs[dims[0]],
                            _raw_idxs[dims[1]],
                        ),
                        dims=result[n].dims,
                    )
            return result


@xr.register_dataset_accessor("iat")
class _Iat_Accessor(_Df_Accessor):
    """
    Multi-dimensional fancy indexing by position.

    Provide the dataset dimensions to index up as keywords, each with
    a value giving an array (one dimensional) of positions to extract.

    All other arguments are keyword-only arguments beginning with an
    underscore.

    Parameters
    ----------
    _name : str, optional
        Only process this variable of this Dataset, and return a DataArray.
    _names : Collection[str], optional
        Only include these variables of this Dataset.
    _load : bool, default False
        Call `load` on the result, which will trigger a compute
        operation if the data underlying this Dataset is in dask,
        otherwise this does nothing.
    _index_name : str, default "index"
        The name to use for the resulting dataset's dimension.
    **idxs : Mapping[str, Any]
        Positions to extract.

    Returns
    -------
    Dataset or DataArray
    """

    def __call__(
        self, *, _name=None, _names=None, _load=False, _index_name=None, **idxs
    ):
        modified_idxs = {}
        raw_idxs = {}

        keep_raw_idxs = False
        is_it_blended = []
        if isinstance(_name, str):
            is_it_blended = [_name]
        elif _names is not None:
            is_it_blended = _names
        for n in is_it_blended:
            try:
                if self._obj.redirection.is_blended(n):
                    keep_raw_idxs = True
            except AttributeError:
                keep_raw_idxs = True

        for k, v in idxs.items():
            target = self._obj.redirection.target(k)
            if target is None:
                modified_idxs[k] = v
                if keep_raw_idxs:
                    raw_idxs[k] = v
            else:
                v_ = np.asarray(v)
                modified_idxs[target] = self._obj[f"_digitized_{target}_of_{k}"][
                    v_
                ].to_numpy()
                if keep_raw_idxs:
                    raw_idxs[target] = v_  # self._obj[k][v_].to_numpy()
        out = self._filter(
            _name=_name,
            _names=_names,
            _load=_load,
            _index_name=_index_name,
            _func="isel",
            _raw_idxs=raw_idxs,
            **modified_idxs,
        )
        return out


@xr.register_dataset_accessor("at")
class _At_Accessor(_Df_Accessor):
    """
    Multi-dimensional fancy indexing by label.

    Provide the dataset dimensions to index up as keywords, each with
    a value giving an array (one dimensional) of labels to extract.

    All other arguments are keyword-only arguments beginning with an
    underscore.

    Parameters
    ----------
    _name : str, optional
        Only process this variable of this Dataset, and return a DataArray.
    _names : Collection[str], optional
        Only include these variables of this Dataset.
    _load : bool, default False
        Call `load` on the result, which will trigger a compute
        operation if the data underlying this Dataset is in dask,
        otherwise this does nothing.
    _index_name : str, default "index"
        The name to use for the resulting dataset's dimension.
    **idxs : Mapping[str, Any]
        Labels to extract.

    Returns
    -------
    Dataset or DataArray
    """

    def __call__(
        self, *, _name=None, _names=None, _load=False, _index_name=None, **idxs
    ):
        modified_idxs = {}
        raw_idxs = {}
        any_redirection = False
        for k, v in idxs.items():
            target = self._obj.redirection.target(k)
            if target is None:
                if any_redirection:
                    raise NotImplementedError(
                        "redirection with `at` must be applied to all "
                        "dimensions or none"
                    )
                raw_idxs[k] = np.asarray(v)
                modified_idxs[k] = v
            else:
                if v.size:
                    upstream = xr.DataArray(np.asarray(v), dims=["index"])
                    downstream = self._obj[k]
                    mapper = {i: j for (j, i) in enumerate(downstream.to_numpy())}
                    offsets = xr.apply_ufunc(np.vectorize(mapper.get), upstream)
                else:
                    offsets = xr.DataArray([], dims=["index"])
                raw_idxs[target] = np.asarray(offsets)
                modified_idxs[target] = self._obj[f"_digitized_{target}_of_{k}"][
                    offsets
                ].to_numpy()
                any_redirection = True
        return self._filter(
            _name=_name,
            _names=_names,
            _load=_load,
            _index_name=_index_name,
            _func="isel" if any_redirection else "sel",
            _raw_idxs=raw_idxs,
            **modified_idxs,
        )


@register_dataset_method
def keep_dims(self, keep_dims, *, errors="raise"):
    """
    Keep only certain dimensions and associated variables from this dataset.

    Parameters
    ----------
    keep_dims : hashable or iterable of hashable
        Dimension or dimensions to keep.
    errors : {"raise", "ignore"}, optional
        If 'raise' (default), raises a ValueError error if any of the
        dimensions passed are not in the dataset. If 'ignore', any given
        dimensions that are in the dataset are dropped and no error is raised.

    Returns
    -------
    obj : Dataset
        The dataset without the given dimensions (or any variables
        containing those dimensions)
    """
    if isinstance(keep_dims, str):
        keep_dims = {keep_dims}
    else:
        keep_dims = set(keep_dims)
    all_dims = set(self.dims)
    if errors == "raise":
        missing_dims = keep_dims - all_dims
        if missing_dims:
            raise ValueError(
                f"{self.__class__} does not contain the dimensions: {missing_dims}"
            )
    return self.drop_dims([i for i in all_dims if i not in keep_dims])



"""Convenience accessor wrappers for xarray objects."""

import xarray as xr


def register_dataset_method(func):
    def wrapper(dataset):
        def f(*args, **kwargs):
            return func(dataset, *args, **kwargs)

        return f

    wrapper.__doc__ = func.__doc__
    return xr.register_dataset_accessor(func.__name__)(wrapper)


def register_dataarray_method(func):
    def wrapper(dataarray):
        def f(*args, **kwargs):
            return func(dataarray, *args, **kwargs)

        return f

    wrapper.__doc__ = func.__doc__
    return xr.register_dataarray_accessor(func.__name__)(wrapper)


def register_dataarray_staticmethod(func):
    return xr.register_dataarray_accessor(func.__name__)(func)


def register_dataset_staticmethod(func):
    return xr.register_dataset_accessor(func.__name__)(func)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np
from rich.table import Table
from scipy.optimize import Bounds, minimize

from ._optional import jax, jnp
from .compiled import jitmethod
from .exceptions import MissingDataError
from .util.simple_attribute import SimpleAttribute

if TYPE_CHECKING:
    from xarray import Dataset


class BucketAccess(ABC):
    @abstractmethod
    def jax_loglike(self, params):
        raise NotImplementedError

    @property
    @abstractmethod
    def pnames(self) -> np.ndarray:
        raise NotImplementedError

    @property
    @abstractmethod
    def pvals(self) -> np.ndarray:
        raise NotImplementedError

    @pvals.setter
    @abstractmethod
    def pvals(self, x):
        raise NotImplementedError

    @property
    @abstractmethod
    def pholdfast(self) -> np.ndarray:
        raise NotImplementedError

    @property
    @abstractmethod
    def pbounds(self) -> Bounds:
        raise NotImplementedError

    @property
    @abstractmethod
    def pstderr(self) -> np.ndarray:
        raise NotImplementedError

    @pstderr.setter
    @abstractmethod
    def pstderr(self, x):
        raise NotImplementedError

    @property
    @abstractmethod
    def parameters(self) -> Dataset:
        raise NotImplementedError

    @abstractmethod
    def add_parameter_array(self, name, values):
        raise NotImplementedError

    @property
    @abstractmethod
    def pretty_table(self):
        raise NotImplementedError


class OptimizeMixin(BucketAccess):
    _cached_loglike_best = None
    log_nans = False

    @abstractmethod
    def jax_loglike(self, params):
        raise NotImplementedError

    @jitmethod
    def jax_d_loglike(self, params):
        return jax.grad(self.jax_loglike)(params)

    @jitmethod
    def jax_hvp_loglike(self, params, vector):
        return jax.grad(lambda x: jnp.vdot(self.jax_d_loglike(x), vector))(params)

    @jitmethod
    def jax_d2_loglike(self, params):
        return jax.jacfwd(jax.jacrev(self.jax_neg_loglike))(params)

    @jitmethod
    def jax_invhess_loglike(self, params):
        hess = self.jax_d2_loglike(params)
        return jnp.linalg.inv(hess)

    def __rich_table(self, params, std_errs=None):
        rich_table = Table()
        rich_table.add_column("Parameter", width=30)
        if std_errs is not None:
            rich_table.add_column("Estimate", width=14)
            rich_table.add_column("Std. Error", width=14)
            rich_table.add_column("t-Stat", width=10)
            rich_table.add_column("Null Val", width=8)
            tstats = (self.pvals - self.pnullvals) / std_errs
            for i in range(params.size):
                rich_table.add_row(
                    str(self.pnames[i]),
                    f"{params[i]: .8g}",
                    " locked" if self.pholdfast[i] else f"{std_errs[i]: .8g}",
                    " locked" if self.pholdfast[i] else f"{tstats[i]: .4g}",
                    f"{self.pnullvals[i]: .2g}",
                )
        else:
            rich_table.add_column("Value", width=14)
            rich_table.add_column("Gradient", width=14)
            rich_table.add_column("Best", width=14)
            if "best" in self.parameters:
                best = self.parameters["best"].to_numpy()
            else:
                best = np.full_like(self.pvals, np.nan)
            for i in range(params.size):
                rich_table.add_row(
                    str(self.pnames[i]),
                    f"{params[i]: .8g}",
                    " locked"
                    if self.pholdfast[i]
                    else f"{self._latest_gradient[i]: .8g}",
                    f"{best[i]: .8g}",
                )
        return rich_table

    def jax_neg_loglike(self, params):
        return -self.jax_loglike(params)

    def __neg_loglike(self, params):
        result = -self.jax_loglike(params)
        self._check_if_best(-result, params)
        if np.isnan(result):
            result = np.inf
        if self.dashboard is not None:
            if self.dashboard.throttle():
                self.dashboard.update_content(
                    loglike=-result,
                    params=self.__rich_table(params),
                    bestloglike=self._cached_loglike_best,
                )
        return result

    def jax_neg_d_loglike(self, *args, **kwargs):
        return -self.jax_d_loglike(*args, **kwargs)

    def __neg_d_loglike(self, *args, **kwargs):
        result = self.jax_neg_d_loglike(*args, **kwargs)
        result = result.at[self.pholdfast != 0].set(0)  # zero out holdfast parameters
        self._latest_gradient = -result
        return result

    def _check_if_best(self, computed_ll, pvalues=None):
        if np.isnan(computed_ll):
            if self.log_nans:
                logging.error("<Detected NaN>")
                if pvalues is None:
                    logging.error(f"{self.pvals}")
                    for z1, z2 in zip(self.pnames, self.pvals):
                        logging.error(f"  {z1:20s} {z2}")
                else:
                    logging.error(f"{pvalues}")
                    for z1, z2 in zip(self.pnames, pvalues):
                        logging.error(f"  {z1:20s} {z2}")
                logging.error("</NaN>")
        elif (
            self._cached_loglike_best is None or computed_ll > self._cached_loglike_best
        ):
            self._cached_loglike_best = computed_ll
            if pvalues is None:
                self.add_parameter_array("best", self.pvals)
            else:
                self.add_parameter_array("best", pvalues)

    dashboard = SimpleAttribute()

    @property
    @abstractmethod
    def dataset(self) -> Dataset:
        raise NotImplementedError

    def jax_maximize_loglike(
        self,
        method="slsqp",
        stderr=False,
        dashboard=True,
        dashboard_update=1.0,
        quiet: bool = False,
        **kwargs,
    ):
        from .model.dashboard import Dashboard

        if self.dataset is None:
            raise MissingDataError("No dataset attached to model")

        if quiet:
            dashboard = False
        self._latest_gradient = np.full_like(self.pvals, np.nan)
        self.dashboard = Dashboard(
            status="[yellow]compiling objective function...",
            params=self.__rich_table(self.pvals),
            show=dashboard,
            throttle=dashboard_update if dashboard else 9999,
        )
        try:
            if method.lower() in {
                "nelder-mead",
                "l-bfgs-b",
                "tnc",
                "slsqp",
                "powell",
                "trust-constr",
            }:
                kwargs["bounds"] = self.pbounds
            self.__neg_loglike(self.pvals)
            self.dashboard.update_content(
                status="[yellow]compiling gradient function..."
            )
            self.__neg_d_loglike(self.pvals)
            self.dashboard.update_content(status="[yellow]optimizing parameters...")
            result = minimize(
                self.__neg_loglike,
                self.pvals,
                jac=self.__neg_d_loglike,
                method=method,
                **kwargs,
            )

            try:
                result["n_cases"] = self.n_cases
            except AttributeError:
                pass
            if hasattr(self, "total_weight"):
                result["total_weight"] = self.total_weight()
            self.pvals = result.x
            if "fun" in result:
                if hasattr(self, "total_weight"):
                    result["logloss"] = result["fun"] / self.total_weight()
                result["loglike"] = -result["fun"]
                del result["fun"]
            if "jac" in result:
                result["jac"] = -result["jac"]
            self.dashboard.update_content(
                loglike=result["loglike"],
                params=self.__rich_table(result.x),
                bestloglike=self._cached_loglike_best,
            )
            if stderr:
                self.dashboard.update_content(
                    status="[yellow]computing standard errors..."
                )
                se, hess, ihess = self.jax_param_cov(result.x)
                result["stderr"] = np.asarray(se)
                hess = np.asarray(hess).copy()
                hess[self.pholdfast.astype(bool), :] = 0
                hess[:, self.pholdfast.astype(bool)] = 0
                ihess = np.asarray(ihess).copy()
                ihess[self.pholdfast.astype(bool), :] = 0
                ihess[:, self.pholdfast.astype(bool)] = 0
                self.add_parameter_array("hess", hess)
                self.add_parameter_array("ihess", ihess)
                self.dashboard.update_content(
                    params=self.__rich_table(result.x, se),
                )
            self.dashboard.update_content(status="[green]optimization complete")
        except Exception as err:
            self.dashboard.update_content(status=f"[bright_red]ERROR [red]{err}")
            raise
        self._most_recent_estimation_result = result
        return result

    def jax_param_cov(self, params):
        hess = self.jax_d2_loglike(params)
        hess = jnp.nan_to_num(hess)
        holds = self.parameters["holdfast"] != 0
        frees = (self.parameters["holdfast"] == 0).astype(jnp.float32).values
        if holds.any():
            hess = hess * frees.reshape(-1, 1)
            hess = hess * frees.reshape(1, -1)
            ihess = jnp.linalg.pinv(hess)
            ihess = ihess * frees.reshape(-1, 1)
            ihess = ihess * frees.reshape(1, -1)
        else:
            ihess = jnp.linalg.inv(hess)
        se = jnp.sqrt(ihess.diagonal())
        self.pstderr = se
        return se, hess, ihess



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import PX, P, X

    m.utility_co[2] = P("ASC_SR2") + P("hhinc#2,3") * X("hhinc")
    m.utility_co[3] = P("ASC_SR3P") + P("hhinc#2,3") * X("hhinc")
    m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
    m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
    m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")

    m.utility_ca = (
        +PX("totcost")
        + P("motorized_time") * X("(altid <= 4) * tottime")
        + P("nonmotorized_time") * X("(altid > 4) * tottime")
        + P("motorized_ovtbydist") * X("(altid <= 4) * ovtt/dist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 7, Diminishing OVTT by Distance"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True)
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_BIKE": -2.675340279013382,
        "ASC_SR2": -2.1871708718901455,
        "ASC_SR3P": -3.5170483581773535,
        "ASC_TRAN": -0.039611848334672385,
        "ASC_WALK": -1.0206078711851112,
        "hhinc#2,3": -0.001383608162986253,
        "hhinc#4": -0.007257729971526785,
        "hhinc#5": -0.01196363695212793,
        "hhinc#6": -0.008182417861122782,
        "motorized_ovtbydist": -0.18145744778558676,
        "motorized_time": -0.04156509908994426,
        "nonmotorized_time": -0.04755705411055601,
        "totcost": -0.004114472584806616,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(expected_x[k], 2e-2), (
            f"{k}, {revealed_x[k] / expected_x[k]}"
        )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import larch as lx

    lx.__version__

    d = lx.examples.MTC()
    m = lx.Model(d, compute_engine="numba")

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    from larch import P, X

    m.utility_ca = (
        +X("totcost/hhinc") * P("costbyincome")
        + X("tottime * (altid <= 4)") * P("motorized_time")
        + X("tottime * (altid >= 5)") * P("nonmotorized_time")
        + X("ovtt/dist * (altid <= 4)") * P("motorized_ovtbydist")
    )

    for a in [4, 5, 6]:
        m.utility_co[a] += X("hhinc") * P(f"hhinc#{a}")

    for i in d["alt_names"][1:3]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_SR")
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    for i in d["alt_names"][3:]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_" + name)
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Zonal",
            "wkcbd.*",
            "wkempden.*",
        ),
        (
            "Household",
            "hhinc.*",
            "vehbywrk.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )
    m.set_cap(25)
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, options={"maxiter": 1000, "ftol": 1e-10})

    m.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import secrets
import tempfile
from pathlib import Path

import numpy as np
import openmatrix
import pandas as pd
import pytest
import xarray as xr
from pytest import approx

import sharrow as sh


def test_dataset_construct_with_zoneids():
    tempdir = tempfile.TemporaryDirectory()
    t = Path(tempdir.name)

    with openmatrix.open_file(t.joinpath("dummy5.omx"), mode="w") as out:
        out.create_carray("/data", "Eye", obj=np.eye(5, dtype=np.float32))
        out.create_carray("/lookup", "Zone", obj=np.asarray([11, 22, 33, 44, 55]))
        shp = np.empty(2, dtype=int)
        shp[0] = 5
        shp[1] = 5
        out.root._v_attrs.SHAPE = shp

    with openmatrix.open_file(t.joinpath("dummy5.omx"), mode="r") as back:
        ds = sh.dataset.from_omx(back, indexes="Zone")

    assert sorted(ds.coords) == ["dtaz", "otaz"]
    assert ds.coords["otaz"].values == approx(np.asarray([11, 22, 33, 44, 55]))
    assert sorted(ds.variables) == ["Eye", "dtaz", "otaz"]
    assert ds["Eye"].data == approx(np.eye(5, dtype=np.float32))

    with openmatrix.open_file(t.joinpath("dummy5.omx"), mode="r") as back:
        ds0 = sh.dataset.from_omx(back, indexes="zero-based")
    assert ds0.coords["otaz"].values == approx(np.asarray([0, 1, 2, 3, 4]))

    with openmatrix.open_file(t.joinpath("dummy5.omx"), mode="r") as back:
        ds1 = sh.dataset.from_omx(back, indexes="one-based")
    assert ds1.coords["otaz"].values == approx(np.asarray([1, 2, 3, 4, 5]))


def test_dataset_categoricals():
    hhs = sh.example_data.get_households()

    def income_cat(i):
        if i < 12500:
            return "LOW"
        elif i < 45000:
            return "MID"
        else:
            return "HIGH"

    hhs["income_grp"] = hhs.income.apply(income_cat).astype(
        pd.CategoricalDtype(["LOW", "MID", "HIGH"], ordered=True)
    )
    assert hhs["income_grp"].dtype == "category"

    hd = sh.dataset.construct(hhs)
    assert hd["income_grp"].dtype == np.int8

    # affirm we can recover categorical and non-categorical data from datarrays
    pd.testing.assert_series_equal(
        hhs["income_grp"], hd.income_grp.single_dim.to_pandas()
    )
    pd.testing.assert_series_equal(hhs["income"], hd.income.single_dim.to_pandas())

    recovered_df = hd.single_dim.to_pandas()
    pd.testing.assert_frame_equal(hhs, recovered_df)


def test_load_with_ignore():
    filename = sh.example_data.get_skims_filename()
    with openmatrix.open_file(filename) as f:
        skims = sh.dataset.from_omx_3d(
            f,
            index_names=("otaz", "dtaz", "time_period"),
            indexes=None,
            time_periods=["EA", "AM", "MD", "PM", "EV"],
            time_period_sep="__",
            max_float_precision=32,
        )
    assert "DRV_COM_WLK_FAR" in skims.variables

    with openmatrix.open_file(filename) as f:
        skims1 = sh.dataset.from_omx_3d(
            f,
            index_names=("otaz", "dtaz", "time_period"),
            indexes=None,
            time_periods=["EA", "AM", "MD", "PM", "EV"],
            time_period_sep="__",
            max_float_precision=32,
            ignore=["DRV_COM_WLK_.*"],
        )
    assert "DRV_COM_WLK_FAR" not in skims1.variables

    with openmatrix.open_file(filename) as f:
        skims2 = sh.dataset.from_omx_3d(
            f,
            index_names=("otaz", "dtaz", "time_period"),
            indexes=None,
            time_periods=["EA", "AM", "MD", "PM", "EV"],
            time_period_sep="__",
            max_float_precision=32,
            ignore="DRV_COM_WLK_.*",
        )
    print(skims2)
    assert "DISTBIKE" in skims2.variables
    assert "DRV_COM_WLK_FAR" not in skims2.variables


def test_deferred_load_to_shared_memory():
    """
    Test of deferred loading of data into shared memory.

    Checks that skim data is loaded correctly into shared memory
    when using the `to_shared_memory` method with `load=False`, followed by
    a call to `reload_from_omx_3d`.
    """
    from sharrow.example_data import get_skims_filename

    skims_filename = get_skims_filename()
    with openmatrix.open_file(skims_filename) as f:
        d0 = sh.dataset.from_omx_3d(
            f,
            index_names=("otaz", "dtaz", "time_period"),
            time_periods=["EA", "AM", "MD", "PM", "EV"],
            max_float_precision=32,
        )
        token = "skims" + secrets.token_hex(5)
        d1 = d0.shm.to_shared_memory(token, mode="r", load=False)
        sh.dataset.reload_from_omx_3d(d1, [skims_filename])
        xr.testing.assert_equal(d0, d1)
        d2 = xr.Dataset.shm.from_shared_memory(token)
        xr.testing.assert_equal(d0, d2)


def test_from_named_objects():
    from sharrow.dataset import from_named_objects

    s1 = pd.Series([1, 4, 9, 16], name="Squares")
    s2 = pd.Series([2, 3, 5, 7, 11], name="Primes")
    i1 = pd.Index([1, 4, 9, 16], name="Squares")
    a1 = xr.DataArray([1, 4, 9, 16], name="Squares")

    for obj in [s1, i1, a1]:
        ds = from_named_objects(obj, s2)
        assert "Squares" in ds.dims
        assert "Primes" in ds.dims
        assert ds.sizes == {"Squares": 4, "Primes": 5}

    with pytest.raises(ValueError):
        from_named_objects([1, 4, 9, 16], s2)


def test_dataarray_iloc():
    arr = xr.DataArray([1, 4, 9, 16, 25, 36], name="Squares", dims="s")

    assert arr.iloc[1] == 4
    xr.testing.assert_equal(arr.iloc[1:], xr.DataArray([4, 9, 16, 25, 36], dims="s"))
    xr.testing.assert_equal(arr.iloc[:2], xr.DataArray([1, 4], dims="s"))
    xr.testing.assert_equal(arr.iloc[2:4], xr.DataArray([9, 16], dims="s"))
    xr.testing.assert_equal(arr.iloc[:-2], xr.DataArray([1, 4, 9, 16], dims="s"))
    xr.testing.assert_equal(arr.iloc[-2:], xr.DataArray([25, 36], dims="s"))

    with pytest.raises(TypeError):
        arr.iloc[1] = 5  # assignment not allowed

    arr2 = xr.DataArray([2, 3, 5, 7, 11], name="Primes", dims="p")
    arr2d = arr * arr2

    with pytest.raises(TypeError):
        _tmp = arr2d.iloc[1]  # not allowed for 2D arrays

    assert arr2d.iloc[dict(s=1, p=2)] == 20

    z = arr2d.iloc[dict(s=slice(1, 2), p=slice(2, 4))]

    xr.testing.assert_equal(z, xr.DataArray([[20, 28]], dims=["s", "p"]))



from __future__ import annotations

import os
import os.path

import numpy as np
import pandas as pd

from ..dataset import DataArray, Dataset

try:
    from .generated import ex
except ImportError:
    ex = None


def _exec_example_n(n, *arg, **kwarg):
    if ex is None:
        raise ImportError("No examples are available")
    return ex[n](*arg, **kwarg)


def example(n, extract="m", estimate=False, output_file=None):
    """Run an example code section (from the documentation) and give the result.

    Parameters
    ----------
    n : int
        The number of the example to reproduce.
    extract : str or Sequence[str]
        The name of the object of the example to extract.  By default `m`
        but it any named object that exists in the example namespace can
        be returned.  Give a sequence of `str` to get multiple objects.
    estimate : bool, default False
        Whether to run the `maximize_loglike` command (and all subsequent
        steps) on the example.
    output_file : str, optional
        If given, check whether this named file exists.  If so,
        return the filename, otherwise run the example code section
        (which should as a side effect create this file).  Then
        check that the file now exists, raising a FileNotFoundError
        if it still does not exist.
    """
    if output_file is not None:
        if os.path.exists(output_file):
            return output_file
        _exec_example_n(n, extract=(), estimate=estimate)
        if os.path.exists(output_file):
            return output_file
        else:
            raise FileNotFoundError(output_file)
    else:
        return _exec_example_n(n, extract=extract, estimate=estimate)


def example_file(filename):
    warehouse_file = os.path.normpath(
        os.path.join(os.path.dirname(__file__), "..", "data_warehouse", filename)
    )
    if os.path.exists(warehouse_file):
        return warehouse_file
    raise FileNotFoundError(
        f"there is no example data file '{warehouse_file}' in data_warehouse"
    )


def MTC(**kwargs):
    ca = pd.read_csv(
        example_file("MTCwork.csv.gz"),
        index_col=("casenum", "altnum"),
        dtype={"chose": np.float32},
    )
    dataset = Dataset.construct.from_idca(
        ca.rename_axis(index=("caseid", "altid")),
        altnames=["DA", "SR2", "SR3+", "Transit", "Bike", "Walk"],
        fill_missing=0,
    )
    dataset["avail"] = DataArray(
        dataset["_avail_"].values, dims=["caseid", "altid"], coords=dataset.coords
    )
    dataset["chose"] = dataset["chose"].fillna(0.0)
    return dataset


def ARTIFICIAL(**kwargs):
    df = pd.read_csv(example_file("artificial_long.csv.gz"))
    df["panel"] = np.ceil(df["id"] / 10).astype(int)
    return Dataset.construct.from_idca(df.set_index(["id", "alt"]))



import logging

import dask.array as da
import numpy as np
import xarray as xr

from .shared_memory import si_units


class ValueDecodeError(ValueError):
    pass


def array_encode(
    x,
    missing_value=None,
    bitwidth=16,
    min_value=None,
    max_value=None,
    scale=None,
    offset=None,
    by_dict=False,
):
    """
    Encode a float DataArray into integers.

    Parameters
    ----------
    x : DataArray
        The original array to be encoded.
    missing_value : Numeric, optional
        If the current array has "missing" values encoded with something
        other that NaN, give that value here.
    bitwidth : {16, 8}
        Number of bits to use in the encoded integers.
    min_value, max_value : Numeric, optional
        Explicitly give the min and max values represented in the array.
        If not given, they are inferred from `x`. It is useful to give
        these values if x does not necessarily include all the values that
        might need to be inserted into `x` later.
    scale : Numeric, optional
        Explicitly give the scaling factor.  This is inferred from the
        min and max values if not provided.
    offset : Numeric, optional
        Explicitly give the offset factor.  This is inferred from the
        min value if not provided.
    by_dict : bool or {8, 16, 32}, optional
        Encode by dictionary, using this bitwidth, or set to `True`
        and give the `bitwidth` argument.  If given, all other
        arguments setting encoding parameters are ignored.

    Returns
    -------
    encoded_values : DataArray
    """
    # if array is already dict encoded in the desired fashion, do nothing
    existing_de = x.attrs.get("digital_encoding", {})
    if by_dict and "dictionary" in existing_de:
        return x
    x = array_decode(x)
    if by_dict:
        if by_dict is True:
            by_dict = bitwidth
        return digitize_by_dictionary(x, bitwidth=by_dict)
    if missing_value is not None:
        legit_values = x.where(x != missing_value)
    else:
        legit_values = x.where(~x.isnull())
    if min_value is None:
        min_value = float(legit_values.min())
    if offset is None:
        offset = min_value
    if max_value is None:
        max_value = float(legit_values.max())
    width = max_value - offset
    n = 1 << (bitwidth - 1)
    if scale is None:
        scale = width / n
    encoded_values = ((x - offset) / scale).astype(f"int{bitwidth}")
    encoded_values = encoded_values.where(x != missing_value, -1)
    encoded_values.attrs["digital_encoding"] = {
        "scale": scale,
        "offset": offset,
        "missing_value": missing_value,
    }
    return encoded_values


def array_decode(x, digital_encoding=None, aux_data=None):
    if digital_encoding is None:
        if "digital_encoding" not in x.attrs:
            return x
        digital_encoding = x.attrs["digital_encoding"]
    dictionary = digital_encoding.get("dictionary", None)
    if dictionary is not None:
        result = x.copy()
        result.data = dictionary[x.to_numpy()]
        result.attrs.pop("digital_encoding", None)
        return result
    offset_source = digital_encoding.get("offset_source", None)
    if offset_source:
        if aux_data is None:
            raise ValueDecodeError(
                "cannot independently decode multivalue DataArray, "
                "provide aux_data or decode from dataset"
            )
        result = aux_data[offset_source].copy()
        result.data = x.to_numpy()[result.data]
        result.attrs.pop("digital_encoding", None)
        return result
    scale = digital_encoding.get("scale", 1)
    offset = digital_encoding.get("offset", 0)
    missing_value = digital_encoding.get("missing_value", None)
    result = x * scale + offset
    if missing_value is not None:
        result = result.where(x >= 0, missing_value)
    result.attrs.pop("digital_encoding", None)
    return result


def smash_bins(values, bincounts, final_width=255, preserve_minmax=True):
    big = 1 << 31
    z = n = bincounts.shape[0]
    if preserve_minmax:
        bincounts[0] = big
        bincounts[-1] = big

    while n > final_width:
        dissolve = np.argmin(bincounts)
        lower_neighbor = dissolve - 1
        while lower_neighbor >= 0 and bincounts[lower_neighbor] == big:
            lower_neighbor -= 1
        upper_neighbor = dissolve + 1
        while upper_neighbor < z and bincounts[upper_neighbor] == big:
            upper_neighbor += 1
        if lower_neighbor >= 0:
            lower_distance = values[dissolve] - values[lower_neighbor]
        else:
            lower_distance = np.inf
        if upper_neighbor < z:
            upper_distance = values[upper_neighbor] - values[dissolve]
        else:
            upper_distance = np.inf
        if lower_distance < upper_distance:
            bincounts[lower_neighbor] += bincounts[dissolve]
        else:
            bincounts[upper_neighbor] += bincounts[dissolve]
        bincounts[dissolve] = big
        n -= 1

    if preserve_minmax:
        bincounts[0] = 1
        bincounts[-1] = 1

    return values[bincounts < big]


def find_bins(values, final_width=255):
    u, b = np.unique(values, return_counts=True)
    return smash_bins(u, b, final_width=final_width)


def digitize_by_dictionary(arr, bitwidth=8):
    result = arr.copy()
    bins = find_bins(arr, final_width=1 << bitwidth)
    try:
        bin_edges = (bins[1:] - bins[:-1]) / 2 + bins[:-1]
    except TypeError:
        # bins are not numeric
        bin_map = {x: n for n, x in enumerate(bins)}
        u, inv = np.unique(arr.data, return_inverse=True)
        result.data = np.array([bin_map.get(x) for x in u])[inv].reshape(arr.shape)
        result.attrs["digital_encoding"] = {
            "dictionary": bins,
        }
        return result
    else:
        try:
            arr_data = arr.data
        except AttributeError:
            pass
        else:
            if isinstance(arr_data, da.Array):
                result.data = da.digitize(arr_data, bin_edges).astype(f"uint{bitwidth}")
                result.attrs["digital_encoding"] = {
                    "dictionary": bins,
                }
                return result
        # fall back to numpy digitize
        result.data = np.digitize(arr, bin_edges).astype(f"uint{bitwidth}")
        result.attrs["digital_encoding"] = {
            "dictionary": bins,
        }
        return result


@xr.register_dataset_accessor("digital_encoding")
class _DigitalEncodings:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def info(self):
        """
        All digital_encoding attributes from Dataset variables.

        Returns
        -------
        dict
        """
        result = {}
        for k in self._obj.variables:
            k_attrs = self._obj._variables[k].attrs
            if "digital_encoding" in k_attrs:
                result[k] = k_attrs["digital_encoding"]
        return result

    def set(self, name, *args, **kwargs):
        """
        Digitally encode one or more variables in this dataset.

        All variables are encoded using the same given parameters.
        To encode various variables differently, make multiple calls
        to this function.

        Parameters
        ----------
        name : str or Collection[str]
            The name(s) of the variable to be encoded.
        missing_value : Numeric, optional
            If the current array has "missing" values encoded with something
            other that NaN, give that value here.
        bitwidth : {16, 8}
            Number of bits to use in the encoded integers.
        min_value, max_value : Numeric, optional
            Explicitly give the min and max values represented in the array.
            If not given, they are inferred from `x`. It is useful to give
            these values if x does not necessarily include all the values that
            might need to be inserted into `x` later.
        scale : Numeric, optional
            Explicitly give the scaling factor.  This is inferred from the
            min and max values if not provided.
        offset : Numeric, optional
            Explicitly give the offset factor.  This is inferred from the
            min value if not provided.
        by_dict : {8, 16, 32}, optional
            Encode by dictionary, using this bitwidth.  If given, all
            arguments other than this and `x` are ignored.
        joint_dict : bool or str, optional
            If given as a string, the variables in `name` will be encoded
            with joint dictionary encoding under this name.  Or simply give
            a `True` value to apply the same with a random unique name.

        Returns
        -------
        Dataset
            A copy of the dataset, with the named variable digitally encoded.
        """
        joint_dict = kwargs.pop("joint_dict", False)
        if joint_dict:
            return multivalue_digitize_by_dictionary(
                self._obj,
                encode_vars=name,
                encoding_name=joint_dict,
            )
        updates = {}
        if isinstance(name, str):
            updates[name] = array_encode(self._obj[name], *args, **kwargs)
        else:
            for n in name:
                updates[n] = array_encode(self._obj[n], *args, **kwargs)
        return self._obj.assign(updates)

    def strip(self, name):
        """
        Digitally decode one or more variables in this dataset.

        Parameters
        ----------
        name : str or Collection[str]
            The name(s) of the variable to be decoded.
        """
        updates = {}
        if isinstance(name, str):
            updates[name] = array_decode(self._obj[name], aux_data=self._obj)
        else:
            for n in name:
                updates[n] = array_decode(self._obj[n], aux_data=self._obj)
        return self._obj.assign(updates)

    def baggage(self, names):
        if isinstance(names, str):
            names = [names]
        if names is None:
            names = list(self._obj)
        bag = set()
        for name in names:
            k_attrs = self._obj._variables[name].attrs
            de = k_attrs.get("digital_encoding", {})
            if "offset_source" in de:
                bag.add(de["offset_source"])
        return bag

    def drop_dims(self, dim_name, *, errors="raise"):
        if isinstance(dim_name, str):
            vars_to_drop = set()
            for k in self._obj.variables:
                k_attrs = self._obj._variables[k].attrs
                if "digital_encoding" in k_attrs:
                    offset_array_name = k_attrs["digital_encoding"].get("offset_source")
                    if dim_name in self._obj._variables[offset_array_name].dims:
                        vars_to_drop.add(k)
                if dim_name in self._obj._variables[k].dims:
                    vars_to_drop.add(k)
            return self._obj.drop_vars(vars_to_drop, errors=errors)
        else:
            obj = self._obj
            for i in dim_name:
                obj = obj.digital_encoding.drop_dims(i, errors=errors)
            return obj


def multivalue_digitize_by_dictionary(ds, encode_vars=None, encoding_name=None):
    """
    Apply a joint dictionary encoding to a collection of Dataset variables.

    Parameters
    ----------
    ds : Dataset
    encode_vars : Collection[str], optional
        The collection of dataset variable names that will be encoded with
        a joint dictionary.
    encoding_name : str, optional
        Use this name for the newly created joint encoding.

    Returns
    -------
    Dataset
    """
    logger = logging.getLogger("sharrow")
    if not isinstance(encoding_name, str):
        i = 0
        while f"joined_{i}" in ds.dims:
            i += 1
        encoding_name = f"joined_{i}"
    if encode_vars is None:
        encode_vars = list(ds)
    if isinstance(encode_vars, str):
        encode_vars = [encode_vars]

    # check each name in encode_vars to make sure it's not already encoded
    # if you want to re-encode first decode
    encode_vars = [
        v
        for v in encode_vars
        if "offset_source" not in ds[v].attrs.get("digital_encoding", {})
    ]
    if len(encode_vars) == 0:
        return ds

    encode_var_dims = ds[encode_vars[0]].dims
    for v in encode_vars[1:]:
        assert encode_var_dims == ds[v].dims, (
            f"dims must match, {encode_var_dims} != {ds[v].dims}"
        )
    logger.info("assembling data stack")
    conjoined = np.stack(
        [array_decode(ds[v].compute(), aux_data=ds) for v in encode_vars], axis=-1
    )
    logger.info("constructing stack view")
    baseshape = conjoined.shape[:-1]
    conjoined = conjoined.reshape([-1, conjoined.shape[-1]])
    voidview = np.ascontiguousarray(conjoined).view(
        np.dtype((np.void, conjoined.dtype.itemsize * conjoined.shape[1]))
    )
    logger.info("finding unique value combinations")
    unique_values, pointers = np.unique(voidview, return_inverse=True)
    pointers = pointers.reshape(baseshape)
    unique_values = unique_values.view(np.dtype(conjoined.dtype)).reshape(
        [-1, len(encode_vars)]
    )
    logger.info("downsampling offsets")
    if unique_values.shape[0] < 1 << 8:
        pointers = pointers.astype(np.uint8)
    elif unique_values.shape[0] < 1 << 16:
        pointers = pointers.astype(np.uint16)
    elif unique_values.shape[0] < 1 << 32:
        pointers = pointers.astype(np.uint32)
    logger.info("formatting output")
    out = ds.drop_vars(encode_vars)

    original_footprint = 0
    encoded_footprint = 0

    out[f"{encoding_name}_offsets"] = xr.DataArray(pointers, dims=encode_var_dims)
    encoded_footprint += pointers.dtype.itemsize * pointers.size

    for n, k in enumerate(encode_vars):
        temp = out[k] = xr.DataArray(
            unique_values[:, n],
            dims=(encoding_name),
            attrs={
                "digital_encoding": {
                    "offset_source": f"{encoding_name}_offsets",
                }
            },
        )
        original_footprint += temp.dtype.itemsize * pointers.size
        encoded_footprint += temp.dtype.itemsize * temp.size

    bytes_saved = original_footprint - encoded_footprint
    savings_ratio = bytes_saved / original_footprint
    logger.info(
        f"multivalue_digitize_by_dictionary {encoding_name} "
        f"saved {si_units(bytes_saved)} {savings_ratio:.1%}"
    )
    return out



def example(extract='m', estimate=False):
    import geopandas as gpd
    import pandas as pd

    import larch as lx

    lx.__version__

    from larch.examples import example_file

    taz_shape = gpd.read_file("zip://" + example_file("exampville_taz.zip"))

    emp = pd.read_csv(example_file("exampville_employment.csv.gz"), index_col="TAZ")

    emp.head()

    from larch.omx import OMX

    skims = OMX(example_file("exampville_skims.omx"), mode="r")
    skims

    hh = pd.read_csv(example_file("exampville_households.csv.gz"))

    hh.head()

    pp = pd.read_csv(example_file("exampville_persons.csv.gz"))

    pp.head()

    tour = pd.read_csv(example_file("exampville_tours.csv.gz"))

    tour.head()
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]

    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import numpy as np
import pytest
import yaml

pytest.importorskip("larch.model.tree")


def test_nesting_arrays():
    nest_yaml = """
    NESTS:
      name: root
      coefficient: coef_nest_root
      alternatives:
          - name: AUTO
            coefficient: coef_nest_AUTO
            alternatives:
                - name: DRIVEALONE
                  coefficient: coef_nest_AUTO_DRIVEALONE
                  alternatives:
                    - DRIVEALONEFREE
                    - DRIVEALONEPAY
                - name: SHAREDRIDE2
                  coefficient: coef_nest_AUTO_SHAREDRIDE2
                  alternatives:
                    - SHARED2FREE
                    - SHARED2PAY
                - name: SHAREDRIDE3
                  coefficient: coef_nest_AUTO_SHAREDRIDE3
                  alternatives:
                    - SHARED3FREE
                    - SHARED3PAY
          - name: NONMOTORIZED
            coefficient: coef_nest_NONMOTORIZED
            alternatives:
              - WALK
              - BIKE
          - name: TRANSIT
            coefficient:  coef_nest_TRANSIT
            alternatives:
                - name: WALKACCESS
                  coefficient: coef_nest_TRANSIT_WALKACCESS
                  alternatives:
                  - WALK_LOC
                  - WALK_LRF
                  - WALK_EXP
                  - WALK_HVY
                  - WALK_COM
                - name: DRIVEACCESS
                  coefficient: coef_nest_TRANSIT_DRIVEACCESS
                  alternatives:
                  - DRIVE_LOC
                  - DRIVE_LRF
                  - DRIVE_EXP
                  - DRIVE_HVY
                  - DRIVE_COM
          - name: RIDEHAIL
            coefficient: coef_nest_RIDEHAIL
            alternatives:
              - TAXI
              - TNC_SINGLE
              - TNC_SHARED
    """
    nests = yaml.safe_load(nest_yaml)["NESTS"]
    alternatives = [
        "DRIVEALONEFREE",
        "DRIVEALONEPAY",
        "SHARED2FREE",
        "SHARED2PAY",
        "SHARED3FREE",
        "SHARED3PAY",
        "WALK",
        "BIKE",
        "WALK_LOC",
        "WALK_LRF",
        "WALK_EXP",
        "WALK_HVY",
        "WALK_COM",
        "DRIVE_LOC",
        "DRIVE_LRF",
        "DRIVE_EXP",
        "DRIVE_HVY",
        "DRIVE_COM",
        "TAXI",
        "TNC_SINGLE",
        "TNC_SHARED",
    ]
    from ..nested_logit import construct_nesting_tree

    tree = construct_nesting_tree(alternatives, nests)
    arrs = tree.as_arrays(
        trim=True,
        parameter_dict={
            "coef_nest_AUTO_DRIVEALONE": 0.3,
            "coef_nest_AUTO_SHAREDRIDE2": 0.4,
            "coef_nest_AUTO_SHAREDRIDE3": 0.4,
            "coef_nest_AUTO": 0.6,
            "coef_nest_NONMOTORIZED": 0.72,
            "coef_nest_TRANSIT": 0.72,
            "coef_nest_TRANSIT_WALKACCESS": 0.5,
            "coef_nest_TRANSIT_DRIVEACCESS": 0.5,
            "coef_nest_RIDEHAIL": 0.36,
        },
    )
    expected = {
        "n_nodes": 31,
        "n_alts": 21,
        "edges_up": np.array(
            [
                21,
                21,
                22,
                22,
                23,
                23,
                24,
                24,
                24,
                25,
                25,
                26,
                26,
                26,
                26,
                26,
                27,
                27,
                27,
                27,
                27,
                28,
                28,
                29,
                29,
                29,
                30,
                30,
                30,
                30,
            ],
            dtype=np.int32,
        ),
        "edges_dn": np.array(
            [
                0,
                1,
                2,
                3,
                4,
                5,
                21,
                22,
                23,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                26,
                27,
                18,
                19,
                20,
                24,
                25,
                28,
                29,
            ],
            dtype=np.int32,
        ),
        "edges_1st": np.array(
            [
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
            ],
            dtype=np.int32,
        ),
        "edges_alloc": np.array(
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            dtype=np.int32,
        ),
        "mu_params": np.array(
            [0.3, 0.4, 0.4, 0.6, 0.72, 0.5, 0.5, 0.72, 0.36, 1.0], dtype=np.float32
        ),
        "start_slots": np.array([0, 2, 4, 6, 9, 11, 16, 21, 23, 26], dtype=np.int32),
        "len_slots": np.array([2, 2, 2, 3, 2, 5, 5, 2, 3, 4], dtype=np.int32),
    }
    for k, v in expected.items():
        assert arrs[k] == pytest.approx(v), f"bad {k!r}"



from __future__ import annotations

import enum

import numpy
import numpy.ma as ma
import pandas

from .. import warning
from . import dictx
from .arraytools import scalarize
from .data_expansion import parse_piece
from .histograms import seems_like_discrete_data, sizable_histogram_figure


def statistics_for_array(
    arr,
    histogram=True,
    varname=None,
    ch_weights=None,
    avail=None,
    dictionary=None,
    flatten=False,
    on_error="ignore",
    **kwargs,
):
    """
    Generate a statistical analysis of an array.

    Parameters
    ----------
    arr : ndarray
            The array to analyze.
    histogram : bool or numeric
            If this evaluates as True, a histogram for the data array
            will be generated. If given as a number, this indicates the
            size and level of detail in the histogram.
    varname : str, optional
            A label for the x-axis of the histogram.  Only appears if
            the size is 3 or greater.  The varname is also used to
            detect piecewise linear segments, when given as
            "piece(label, lowerbound, upperbound)". Note this function
            does not *process* the array to extract a linear segment,
            this just flags the histogram generator to know that this
            is a linear segment.
    ch_weights : ndarray, optional
            If given, this indicates the choice-weights of the observations.
            Must be the same shape as `arr`. Useful to analyze whether the
            distribution of the data appears different for chosen
            alternatives than for all alternatives.
    avail : ndarray, optional
            If given, this indicates the availability of the observations.
            Must be the same shape as `arr`. If missing, all finite values
            in `arr` are considered available.
    dictionary : dict, optional
            A dictionary of {value: str} pairs, which give labels to use on
            the histogram instead of the code values stored in `arr`.
    flatten : bool, default False
            Indicates if `arr` (and `ch_weights` if given) should be
            flattened before processing.
    on_error : {'ignore', 'raise'}
            How to handle exceptions encountered while generating statistics.

    Other keyword parameters are passed through to
    :meth:`sizable_histogram_figure`.

    Returns
    -------
    dictx

    """
    a = arr

    if flatten:
        a = a.reshape(-1)
        if ch_weights is not None:
            ch_weights = ch_weights.reshape(-1)

    stats = dictx()

    if varname is not None:
        stats.description = varname

    base_def, lowbound, highbound, piece_exactly = parse_piece(varname)

    try:
        _nans = numpy.isnan(a)
    except TypeError:
        can_nan = False
    else:
        can_nan = True

    a_masked_is_category = None

    if avail is None and can_nan:
        a_masked = ma.masked_array(a, mask=~numpy.isfinite(a))
        stats.n = scalarize(numpy.isfinite(a).sum(axis=0))
    elif avail is not None:
        a_masked = ma.masked_array(a, mask=~avail)
        stats.n = scalarize((avail != 0).sum(axis=0))
    elif isinstance(a, pandas.Series) and isinstance(a.dtype, pandas.CategoricalDtype):
        a_masked = ma.masked_array(a, mask=pandas.isnull(a))
        a_masked_is_category = True
        stats.n = scalarize((~pandas.isnull(a)).sum(axis=0))
    else:
        a_masked = a
        stats.n = scalarize(a.shape[0])

    if can_nan:
        stats.minimum = scalarize(numpy.nanmin(a_masked, axis=0))
        stats.maximum = scalarize(numpy.nanmax(a_masked, axis=0))
        with warning.ignore_warnings():
            if ma.is_masked:
                stats.median = scalarize(ma.median(a_masked, axis=0))
            else:
                stats.median = scalarize(numpy.nanmedian(a_masked, axis=0))

    if ch_weights is not None:
        a_ch_weighted = ma.masked_array(a, mask=(ch_weights <= 0))
        stats.n_chosen = scalarize(a_ch_weighted.count(axis=0))
        if can_nan:
            stats.minimum_chosen = scalarize(numpy.nanmin(a_ch_weighted, axis=0))
            stats.maximum_chosen = scalarize(numpy.nanmax(a_ch_weighted, axis=0))
            with warning.ignore_warnings():
                stats.median_chosen = scalarize(numpy.nanmedian(a_ch_weighted, axis=0))

    if ch_weights is not None:
        if avail is None and can_nan:
            ch_weightsx = ma.masked_array(ch_weights, mask=~numpy.isfinite(a))
        elif avail is not None:
            ch_weightsx = ma.masked_array(ch_weights, mask=~avail)
        elif isinstance(a, pandas.Series) and isinstance(
            a.dtype, pandas.CategoricalDtype
        ):
            ch_weightsx = ma.masked_array(ch_weights, mask=pandas.isnull(a))
        else:
            ch_weightsx = ch_weights
    else:
        ch_weightsx = None

    if a_masked_is_category is None:
        try:
            a_masked_is_category = isinstance(a_masked.dtype, pandas.CategoricalDtype)
        except Exception:
            a_masked_is_category = False

    if (can_nan and histogram) or a_masked_is_category:
        with warning.ignore_warnings():
            discrete_ = kwargs.pop("discrete", None)
            try:
                if a_masked_is_category:
                    discrete_ = True
            except TypeError:
                pass
            if discrete_ is None:
                discrete_ = seems_like_discrete_data(a_masked, dictionary)

            if lowbound is not None or highbound is not None:
                stats.histogram = sizable_histogram_figure(
                    a_masked,
                    sizer=histogram,
                    title=None,
                    xlabel=base_def,
                    ylabel="Frequency",
                    ch_weights=ch_weightsx,
                    piecerange=(lowbound, highbound),
                    **kwargs,
                )
            else:
                stats.histogram = sizable_histogram_figure(
                    a_masked,
                    sizer=histogram,
                    title=None,
                    xlabel=varname,
                    ylabel="Frequency",
                    ch_weights=ch_weightsx,
                    discrete=discrete_,
                    on_error=on_error,
                    **kwargs,
                )
    if can_nan:
        stats.mean = scalarize(numpy.mean(a_masked, axis=0))
        stats.stdev = scalarize(numpy.std(a_masked, axis=0))
        stats.zeros = scalarize(numpy.sum(numpy.logical_not(a_masked), axis=0))
        with warning.ignore_warnings():
            stats.positives = scalarize(numpy.sum(a_masked > 0, axis=0))
            stats.negatives = scalarize(numpy.sum(a_masked < 0, axis=0))
        n_nans = scalarize(numpy.sum(numpy.isnan(a), axis=0))
        if numpy.any(n_nans):
            stats.nans = n_nans
        n_infs = scalarize(numpy.sum(numpy.isinf(a), axis=0))
        if numpy.any(n_infs):
            stats.infs = n_infs
        a_masked.mask |= a_masked == 0
        stats.nonzero_minimum = scalarize(numpy.min(a_masked, axis=0))
        stats.nonzero_maximum = scalarize(numpy.max(a_masked, axis=0))
        stats.nonzero_mean = scalarize(numpy.mean(a_masked, axis=0))
        stats.nonzero_stdev = scalarize(numpy.std(a_masked, axis=0))
    elif a_masked_is_category:
        stats.nulls = pandas.isnull(a).sum()
        the_mode = tuple(a.mode().values)
        if len(the_mode) == 1:
            the_mode = the_mode[0]
        stats.mode = the_mode

    if dictionary is not None:
        stats.dictionary = dictionary

    return stats


def statistics_for_array5(arr, histogram=5, *args, **kwargs):
    """
    Generate a statistical analysis of an array.

    Parameters
    ----------
    arr : ndarray
            The array to analyze.
    histogram : bool or numeric, default 5
            If this evaluates as True, a histogram for the data array
            will be generated. If given as a number, this indicates
            the size and level of detail in the histogram.
    varname : str, optional
            A label for the x-axis of the histogram.  Only appears if
            the size is 3 or greater.  The varname is also used to
            detect piecewise linear segments, when given as
            "piece(label, lowerbound, upperbound)". Note this function
            does not *process* the array to extract a linear segment,
            this just flags the histogram generator to know that this
            is a linear segment.
    ch_weights : ndarray, optional
            If given, this indicates the choice-weights of the
            observations.  Must be the same shape as `arr`. Useful to
            analyze whether the distribution of the data appears
            different for chosen alternatives than for all alternatives.
    avail : ndarray, optional
            If given, this indicates the availability of the observations.
            Must be the same shape as `arr`. If missing, all finite values
            in `arr` are considered available.
    dictionary : dict, optional
            A dictionary of {value: str} pairs, which give labels to use
            on the histogram instead of the code values stored in `arr`.
    flatten : bool, default False
            Indicates if `arr` (and `ch_weights` if given) should be
            flattened before processing.

    Other keyword parameters are passed through to
    :meth:`sizable_histogram_figure`.

    Returns
    -------
    dictx

    """
    return statistics_for_array(arr, *args, histogram=histogram, **kwargs)


def statistics_for_dataframe(df, histogram=True, ch_weights=None, avail=None, **kwargs):
    s = {}
    for col in df.columns:
        if isinstance(df.dtypes[col], pandas.CategoricalDtype):
            s[col] = statistics_for_array(
                df[col],
                varname=col,
                histogram=histogram,
                ch_weights=ch_weights,
                avail=avail,
                **kwargs,
            )
        elif numpy.issubdtype(df.dtypes[col], numpy.floating):
            s[col] = statistics_for_array(
                df[col],
                varname=col,
                histogram=histogram,
                ch_weights=ch_weights,
                avail=avail,
                **kwargs,
            )
        elif numpy.issubdtype(df.dtypes[col], numpy.integer):
            s[col] = statistics_for_array(
                df[col],
                varname=col,
                histogram=histogram,
                ch_weights=ch_weights,
                avail=avail,
                **kwargs,
            )
        else:
            s[col] = statistics_for_array(
                df[col],
                varname=col,
                histogram=histogram,
                ch_weights=ch_weights,
                avail=avail,
                **kwargs,
            )

    from .dataframe import DataFrameViewer

    result = DataFrameViewer.from_dict(s, orient="index")
    if "nans" in result.columns:
        result["nans"].fillna(0, inplace=True)
    try:
        result = result.drop("description", axis=1)
    except Exception:
        pass
    return result


def uniques(s, dictionary=None):
    """
    Count the frequency for each unique value in a pandas.Series.

    Parameters
    ----------
    s : pandas.Series
            The values to count.
    dictionary : Mapping or Enum, optional
            A mapping converting the actual values in the series
            to a more meaningful value (e.g. changes code numbers
            to names).  If an Enum class is passed, the values
            are treated as keys, and the names as values.

    Returns
    -------
    pandas.Series
    """
    action = s
    len_action = len(action)
    try:
        action = action[~numpy.isnan(action)]
    except TypeError:
        num_nan = 0
    else:
        num_nan = len_action - len(action)
    x = numpy.unique(action, return_counts=True)
    if dictionary is None:
        y = pandas.Series(x[1], x[0])
    else:
        if isinstance(dictionary, enum.EnumMeta):
            dictionary = {v: k for k, v in dictionary.__members__.items()}
        y = pandas.Series(x[1], [dictionary.get(j, j) for j in x[0]])
    if num_nan:
        y[numpy.nan] = num_nan
    return y


def invmap(s, mapping):
    """
    Apply an inverse map.

    When the values in `s` are the values of a Mapping,
    but you want them to be the keys instead.  Note
    that values need not be unique, so an arbitrary
    key with the correct value will result for non-unique
    inverse mappings.

    Parameters
    ----------
    s : array-like
    mapping : Mapping or Enum, optional
            A mapping. If an Enum class is passed, the
            __members__ is used.

    Returns
    -------
    array-like
    """
    if isinstance(mapping, enum.EnumMeta):
        mapping = mapping.__members__
    inv_map = {v: k for k, v in mapping.items()}
    return s.map(inv_map)



import numpy as np
import pandas as pd
import xarray as xr


def _iat(source, *, _names=None, _load=False, _index_name=None, **idxs):
    loaders = {}
    if _index_name is None:
        _index_name = "index"
    for k, v in idxs.items():
        loaders[k] = xr.DataArray(v, dims=[_index_name])
    if _names:
        ds = source[_names]
    else:
        ds = source
    if _load:
        ds = ds._load()
    return ds.isel(**loaders)


def _at(source, *, _names=None, _load=False, _index_name=None, **idxs):
    loaders = {}
    if _index_name is None:
        _index_name = "index"
    for k, v in idxs.items():
        loaders[k] = xr.DataArray(v, dims=[_index_name])
    if _names:
        ds = source[_names]
    else:
        ds = source
    if _load:
        ds = ds._load()
    return ds.sel(**loaders)


def gather(source, indexes):
    """
    Extract values by label on the coordinates indicated by columns of a DataFrame.

    Parameters
    ----------
    source : xarray.DataArray or xarray.Dataset
        The source of the values to extract.
    indexes : Mapping[str, array-like]
        The keys of `indexes` (if given as a dataframe, the column names)
        should match the named dimensions of `source`.  The resulting extracted
        data will have a shape one row per row of `df`, and columns matching
        the data variables in `source`, and each value is looked up by the labels.

    Returns
    -------
    pd.DataFrame
    """
    result = _at(source, **indexes).reset_coords(drop=True)
    return result


def igather(source, positions):
    """
    Extract values by position on the coordinates indicated by columns of a DataFrame.

    Parameters
    ----------
    source : xarray.DataArray or xarray.Dataset
    positions : pd.DataFrame or Mapping[str, array-like]
        The columns (or keys) of `df` should match the named dimensions of
        this Dataset.  The resulting extracted DataFrame will have one row
        per row of `df`, columns matching the data variables in this dataset,
        and each value is looked up by the positions.

    Returns
    -------
    pd.DataFrame
    """
    result = _iat(source, **positions).reset_coords(drop=True)
    return result


class DatasetWrapper:
    def __init__(self, dataset, orig_key, dest_key, time_key=None):
        """
        Emulate ActivitySim's SkimWrapper.

        Parameters
        ----------
        skim_dict: SkimDict

        orig_key: str
            name of column in dataframe to use as implicit orig for lookups
        dest_key: str
            name of column in dataframe to use as implicit dest for lookups
        """
        self.dataset = dataset
        self.orig_key = orig_key
        self.dest_key = dest_key
        self.time_key = time_key
        self.df = None

    def set_df(self, df):
        """
        Set the dataframe.

        Parameters
        ----------
        df : DataFrame
            The dataframe which contains the origin and destination ids

        Returns
        -------
        self (to facilitate chaining)
        """
        assert self.orig_key in df, (
            f"orig_key '{self.orig_key}' not in df columns: {list(df.columns)}"
        )
        assert self.dest_key in df, (
            f"dest_key '{self.dest_key}' not in df columns: {list(df.columns)}"
        )
        if self.time_key:
            assert self.time_key in df, (
                f"time_key '{self.time_key}' not in df columns: {list(df.columns)}"
            )
        self.df = df

        # TODO allow non-1 offsets
        positions = {
            "otaz": df[self.orig_key] - 1,
            "dtaz": df[self.dest_key] - 1,
        }
        if self.time_key:
            time_map = {j: i for i, j in enumerate(self.dataset.indexes["time_period"])}
            positions["time_period"] = pd.Series(
                np.vectorize(time_map.get)(df[self.time_key]),
                index=df.index,
            )
        self.positions = pd.DataFrame(positions)

        return self

    def lookup(self, key, reverse=False):
        """
        Generally not called by the user - use __getitem__ instead.

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        od : bool (optional)
            od=True means lookup standard origin-destination skim value
            od=False means lookup destination-origin skim value

        Returns
        -------
        impedances: pd.Series
            A Series of impedances which are elements of the Skim object and
            with the same index as df
        """
        assert self.df is not None, "Call set_df first"
        if reverse:
            x = self.positions.rename(columns={"otaz": "dtaz", "dtaz": "otaz"})
        else:
            x = self.positions

        # Return a series, consistent with ActivitySim SkimWrapper
        return igather(self.dataset[key], x).to_series()

    def __getitem__(self, key):
        """
        Get the lookup for an available skim object.

        The `df` and orig/dest and column names are implicit.

        Parameters
        ----------
        key : hashable
             The key (identifier) for the skim object

        Returns
        -------
        impedances: pd.Series with the same index as df
            A Series of impedances values from the single Skim with specified key,
            indexed byt orig/dest pair
        """
        return self.lookup(key)



from xarray import DataArray

from . import dataset, example_data, selectors, shared_memory, sparse
from ._infer_version import __version__, __version_tuple__
from .dataset import Dataset
from .datastore import DataStore
from .digital_encoding import array_decode, array_encode
from .flows import CacheMissWarning, Flow
from .relationships import DataTree, Relationship
from .table import Table, concat_tables

__all__ = [
    "__version__",
    "__version_tuple__",
    "DataArray",
    "Dataset",
    "DataStore",
    "DataTree",
    "Relationship",
    "Table",
    "CacheMissWarning",
    "Flow",
    "example_data",
    "array_decode",
    "array_encode",
    "concat_tables",
    "dataset",
    "selectors",
    "shared_memory",
    "sparse",
]



from __future__ import annotations

from pytest import approx

from larch import PX, P, X
from larch.model.linear import DataRef


def test_P():
    assert P("one") == "one"


def test_X():
    # Add test cases for X function here
    assert X(2) == "2"
    assert X(3.14) == "3.14"
    assert X("hello") == "hello"


def test_PX():
    # Add test cases for PX function here
    assert PX("hello") == P.hello * X.hello


def test_linear_components():
    # Add test cases for LinearFunction class here
    f = P.Aaa * X.Bbb + P.Ccc * X.Ddd
    assert f[0].param == "Aaa"
    assert f[0].data == "Bbb"
    assert f[1].param == "Ccc"
    assert f[1].data == "Ddd"
    assert len(f) == 2


def test_linear_component_math():
    # Add test cases for LinearFunction class here
    f = P.Aaa * X.Bbb * 3
    assert f.param == "Aaa"
    assert f.data == "Bbb"
    assert f.scale == 3

    f = P.Aaa * (X.Bbb * 3)
    assert f.param == "Aaa"
    assert f.data == "Bbb"
    assert f.scale == 3

    f = P.Aaa * (3 * X.Bbb)
    assert f.param == "Aaa"
    assert f.data == "Bbb"
    assert f.scale == 3

    f = P.Aaa * (X.Bbb * 3.1)
    assert f.param == "Aaa"
    assert f.data == "Bbb"
    assert f.scale == 3.1

    f = P.Aaa * (3.1 * X.Bbb)
    assert f.param == "Aaa"
    assert f.data == "Bbb"
    assert f.scale == 3.1

    f = P.Aaa / X.Bbb
    assert f.param == "Aaa"
    assert f.data == "1/Bbb"
    assert f.scale == 1

    f = 1 / X.Bbb
    assert f == "1/Bbb"
    assert f.__class__ is DataRef

    f = 1.1 / X.Bbb
    assert f.param == "_"
    assert f.data == "1/Bbb"
    assert f.scale == 1.1

    f = P.Aaa * (X.Bbb / X.Ccc)
    assert f.param == "Aaa"
    assert f.data == "Bbb/Ccc"
    assert f.scale == 1

    f = (P.Aaa * X.Bbb) / X.Ccc
    assert f.param == "Aaa"
    assert f.data == "Bbb/Ccc"
    assert f.scale == 1

    f = (P.Aaa * X.Bbb + P.Ccc * X.Ddd) * X.Eee
    assert len(f) == 2
    assert f[0].param == "Aaa"
    assert f[0].data == "Bbb*Eee"
    assert f[1].param == "Ccc"
    assert f[1].data == "Ddd*Eee"

    f = (P.Aaa * X.Bbb + P.Ccc * X.Ddd) / X.Eee
    assert len(f) == 2
    assert f[0].param == "Aaa"
    assert f[0].data == "Bbb/Eee"
    assert f[1].param == "Ccc"
    assert f[1].data == "Ddd/Eee"

    f = (P.Aaa * X.Bbb + P.Ccc * X.Ddd) * 1.1
    assert len(f) == 2
    assert f[0].param == "Aaa"
    assert f[0].data == "Bbb"
    assert f[0].scale == 1.1
    assert f[1].param == "Ccc"
    assert f[1].data == "Ddd"
    assert f[1].scale == 1.1

    f = (P.Aaa * X.Bbb + P.Ccc * X.Ddd) / 1.1
    assert len(f) == 2
    assert f[0].param == "Aaa"
    assert f[0].data == "Bbb"
    assert f[0].scale == approx(1.0 / 1.1)
    assert f[1].param == "Ccc"
    assert f[1].data == "Ddd"
    assert f[1].scale == approx(1.0 / 1.1)



from __future__ import annotations

import pytest

import larch
from larch import PX, P, X


def test_old_commands():
    d = larch.examples.MTC()

    with pytest.raises(NotImplementedError):
        m = larch.Model(dataservice=d)

    m = larch.Model(datatree=d)

    m.utility_co[2] = P("ASC_SR2") + P("hhinc#2") * X("hhinc")
    m.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
    m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
    m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
    m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
    m.utility_ca = PX("tottime") + PX("totcost")

    with pytest.raises(NotImplementedError):
        m.availability_var = "_avail_"

    m.availability_ca_var = "_avail_"
    m.choice_ca_var = "_choice_"
    m.title = "MTC Example 1 (Simple MNL)"

    with pytest.warns(DeprecationWarning):
        m.load_data()

    with pytest.raises(ValueError):
        m.maximize_loglike()

    m.choice_ca_var = "chose"
    result = m.maximize_loglike()
    assert result.loglike == pytest.approx(-3626.18625551293)



from __future__ import annotations

from ._optional import jax


def reset_compiled_methods(obj):
    precompiled_funcs = [i for i in obj.__dict__ if i.startswith("_precompiled_")]
    for i in precompiled_funcs:
        delattr(obj, i)


class compiledmethod:
    """Decorator for a class method that returns a compiled function."""

    def __init__(self, compiler):
        # self : compiledmethod
        # compiler : the class method being decorated
        self.compiler = compiler
        self.docstring = compiler.__doc__

    def __set_name__(self, owner, name):
        # self : compiledmethod
        # owner : parent class that will have `self` as a member
        # name : the name of the attribute that `self` will be
        self.public_name = name
        self.private_name = "_precompiled_" + name

    def __get__(self, obj, objtype=None):
        # self : compiledmethod
        # obj : instance of parent class that has `self` as a member
        # objtype : class of `obj`
        unmangle = getattr(obj, "unmangle", None)
        if unmangle is not None:
            unmangle()
        result = getattr(obj, self.private_name, None)
        if result is None:
            result = self.compiler(obj)
            result.__doc__ = self.docstring
            setattr(obj, self.private_name, result)
        return result

    def __set__(self, obj, value):
        # self : compiledmethod
        # obj : instance of parent class that has `self` as a member
        # value : the new value that is trying to be assigned
        raise AttributeError(f"can't set {self.public_name}")

    def __delete__(self, obj):
        # self : compiledmethod
        # obj : instance of parent class that has `self` as a member
        setattr(obj, self.private_name, None)


class jitmethod:
    """Decorator for a class method that returns a compiled function."""

    def __init__(self, wrapped_method=None, **kwargs):
        """
        Initialize a jit compile wrapper.

        Parameters
        ----------
        wrapped_method : Callable
            The class method being decorated.
        """
        if wrapped_method is None:
            self.jit_kwargs = kwargs
        else:
            self.jit_kwargs = {}
            self.wrapped_method = wrapped_method
            self.docstring = wrapped_method.__doc__

    def __call__(self, wrapped_method):
        self.wrapped_method = wrapped_method
        self.docstring = wrapped_method.__doc__
        return self

    def __set_name__(self, owner, name):
        """
        Triggered on assignment to a class name.

        Parameters
        ----------
        owner : Any
            Parent class that will have `self` as a member.
        name : str
            The name of the attribute to which self is being assigned.
        """
        self.public_name = name
        self.private_name = "_precompiled_" + name

    def __get__(self, obj, objtype=None):
        """
        Access to jit'ed method.

        Parameters
        ----------
        obj : Any
            Instance of parent class that has `self` as a member.
        objtype : class
            Class of `obj`.
        """
        unmangle = getattr(obj, "unmangle", None)
        if unmangle is not None:
            unmangle()
        result = getattr(obj, self.private_name, None)
        if result is None:

            def func(*args, **kwargs):
                return self.wrapped_method(obj, *args, **kwargs)

            result = jax.jit(func, **self.jit_kwargs)
            result.__doc__ = self.docstring
            setattr(obj, self.private_name, result)
        return result

    def __set__(self, obj, value):
        raise AttributeError(f"can't set {self.public_name} is is a jitmethod")

    def __delete__(self, obj):
        """
        Clear precompiled jit'ed method.

        Parameters
        ----------
        obj : Any
            Instance of parent class that has `self` as a member.
        """
        setattr(obj, self.private_name, None)



import os
import sys
from pathlib import Path

from ruamel.yaml import YAML

import larch

config_file = Path(
    os.path.join(
        os.path.dirname(__file__),
        "..",
        "_config.yml",
    )
)

if len(sys.argv) >= 2:
    title = sys.argv[1]
else:
    title = larch.__version__
yaml = YAML(typ="rt")  # default, if not specfied, is 'rt' (round-trip)
content = yaml.load(config_file)
content["title"] = title
content["html"]["extra_footer"] = f"<p>Larch {title}</p>"
yaml.dump(content, config_file)
